"""Unit of work"""

from __future__ import annotations

import abc
import enum
from collections.abc import Iterator
from types import TracebackType
from typing import Any, Generic, Optional, TypeVar

from jeepito.domain.model import Message
from jeepito.service._async.repository import (
    AsyncAbstractRepository,
    AsyncEventstoreAbstractRepository,
    AsyncSinkholeEventstoreRepository,
)


class TransactionError(RuntimeError):
    """A runtime error raised if the transaction lifetime is inappropriate."""


class TransactionStatus(enum.Enum):
    """Transaction status used to ensure transaction lifetime."""

    running = "running"
    rolledback = "rolledback"
    committed = "committed"
    closed = "closed"


TRepositories = TypeVar("TRepositories", bound=AsyncAbstractRepository[Any])


class AsyncUnitOfWorkTransaction(Generic[TRepositories]):
    uow: AsyncAbstractUnitOfWork[TRepositories]
    status: TransactionStatus

    def __init__(self, uow: AsyncAbstractUnitOfWork[TRepositories]) -> None:
        self.status = TransactionStatus.running
        self.uow = uow

    def __getattr__(self, name: str) -> TRepositories:
        return getattr(self.uow, name)

    @property
    def eventstore(self) -> AsyncEventstoreAbstractRepository:
        return self.uow.eventstore

    async def commit(self) -> None:
        if self.status != TransactionStatus.running:
            raise TransactionError(f"Transaction already closed ({self.status.value}).")
        await self.uow.commit()
        self.status = TransactionStatus.committed

    async def rollback(self) -> None:
        await self.uow.rollback()
        self.status = TransactionStatus.rolledback

    async def __aenter__(self) -> AsyncUnitOfWorkTransaction[TRepositories]:
        if self.status != TransactionStatus.running:
            raise TransactionError("Invalid transaction status.")
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        """Rollback in case of exception."""
        if exc:
            await self.rollback()
            return
        if self.status == TransactionStatus.closed:
            raise TransactionError("Transaction is closed.")
        if self.status == TransactionStatus.running:
            raise TransactionError(
                "Transaction must be explicitly close. Missing commit/rollback call."
            )
        if self.status == TransactionStatus.committed:
            await self.uow.eventstore.publish_eventstream()
        self.status = TransactionStatus.closed


class AsyncAbstractUnitOfWork(abc.ABC, Generic[TRepositories]):
    """
    Abstract unit of work.

    To implement a unit of work, the :meth:`AsyncAbstractUnitOfWork.commit` and
    :meth:`AsyncAbstractUnitOfWork.rollback` has to be defined, and some repositories
    has to be declared has attributes.
    """

    eventstore: AsyncEventstoreAbstractRepository = AsyncSinkholeEventstoreRepository()

    def collect_new_events(self) -> Iterator[Message[Any]]:
        for repo in self._iter_repositories():
            while repo.seen:
                model = repo.seen.pop(0)
                while model.messages:
                    yield model.messages.pop(0)

    def _iter_repositories(
        self,
    ) -> Iterator[AsyncAbstractRepository[Any]]:
        for member_name in self.__dict__.keys():
            member = getattr(self, member_name)
            if isinstance(member, AsyncAbstractRepository):
                yield member

    async def __aenter__(self) -> AsyncUnitOfWorkTransaction[TRepositories]:
        self.__transaction = AsyncUnitOfWorkTransaction(self)
        await self.__transaction.__aenter__()
        return self.__transaction

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        # AsyncUnitOfWorkTransaction is making the thing
        await self.__transaction.__aexit__(exc_type, exc, tb)

    @abc.abstractmethod
    async def commit(self) -> None:
        """Commit the transation."""

    @abc.abstractmethod
    async def rollback(self) -> None:
        """Rollback the transation."""



"""
Message base classes.

`Command` and `Event` are two types used to handle changes in the model.

"""

from collections.abc import MutableSequence
from datetime import datetime
from typing import Any, Generic, TypeVar
from uuid import uuid1

from pydantic import BaseModel, Field


def generate_id() -> str:
    """Generate a unique identifier."""
    return str(uuid1())


class Metadata(BaseModel):
    name: str = Field(...)
    """Name of the schema."""
    schema_version: int = Field(...)
    """Version of the schema."""
    published: bool = Field(default=False)
    """Publish the event to an eventstream."""


TMetadata = TypeVar("TMetadata", bound=Metadata)


class Message(BaseModel, Generic[TMetadata]):
    """Base class for messaging."""

    message_id: str = Field(default_factory=generate_id)
    """Unique identifier of the message."""
    created_at: datetime = Field(default_factory=datetime.utcnow)
    """
    Timestamp of the message.

    All messages are kept in order for observability, debug and event replay.
    """
    metadata: TMetadata
    """
    Define extra fields used at serialization.

    While serializing the message, a name and version must be defined to properly
    defined the message. Event if the class is renamed, those constants must be kept
    identically over the time in the codebase.

    metadata are defined statically at the definition of the message.
    """

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Message):
            return False
        slf = self.model_dump(exclude={"message_id", "created_at"})
        otr = other.model_dump(exclude={"message_id", "created_at"})
        return slf == otr


class Command(Message[TMetadata]):
    """Baseclass for message of type command."""


class Event(Message[TMetadata]):
    """Baseclass for message of type event."""


class Model(BaseModel, Generic[TMetadata]):
    """Base class for model."""

    messages: MutableSequence[Message[TMetadata]] = Field(default_factory=list)
    """
    List of messages consumed by the unit of work to mutate the repository.

    Those message are ephemeral, published by event handler and consumed
    by the unit of work during the process of an original command.
    """

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Model):
            return False
        slf = self.model_dump(exclude={"messages"})
        otr = other.model_dump(exclude={"messages"})
        return slf == otr



"""
Propagate commands and events to every registered handles.

"""

import logging
from collections.abc import Callable, Coroutine
from typing import Any, TypeVar

from jeepito.domain.model import Message

from .service._async.unit_of_work import AsyncAbstractUnitOfWork
from .service._sync.unit_of_work import SyncAbstractUnitOfWork

log = logging.getLogger(__name__)

TAsyncUow = TypeVar("TAsyncUow", bound=AsyncAbstractUnitOfWork[Any])
TSyncUow = TypeVar("TSyncUow", bound=SyncAbstractUnitOfWork[Any])
TMessage = TypeVar("TMessage", bound=Message[Any])

AsyncMessageHandler = Callable[[TMessage, TAsyncUow], Coroutine[Any, Any, Any]]
SyncMessageHandler = Callable[[TMessage, TSyncUow], Any]



import abc
from collections.abc import Mapping
from typing import Any

from jeepito.domain.model import Message


class AbstractMessageSerializer(abc.ABC):
    """The message serializer take the message and return it with python native types,
    that are serializable for a transport.
    """

    @abc.abstractmethod
    def serialize_message(self, message: Message[Any]) -> Mapping[str, Any]:
        """Publish a message to the eventstream."""


class MessageSerializer(AbstractMessageSerializer):
    """Default message serializer"""

    def serialize_message(self, message: Message[Any]) -> Mapping[str, Any]:
        """Publish a message to the eventstream."""
        return {
            "id": message.message_id,
            "created_at": message.created_at.isoformat(),
            "type": f"{message.metadata.name}_v{message.metadata.schema_version}",
            "payload": message.model_dump_json(
                exclude={"message_id", "created_at", "metadata"}
            ),
        }



"""
Repositories are used to fetch and store domain models.

Anstract repositories derived the :class:`jeepito.AsyncAbstractRepository`
class to declare every models interface such as CRUD operations,
and then concrete models implements those abstract methods for a given
storage.
"""

import abc
from collections.abc import MutableSequence
from typing import Any, Generic, Optional, TypeVar

from jeepito.domain.model import Message, Model
from jeepito.service._sync.eventstream import SyncEventstreamPublisher

TModel_contra = TypeVar("TModel_contra", bound=Model[Any], contravariant=True)


class SyncAbstractRepository(abc.ABC, Generic[TModel_contra]):
    """Abstract Base Classe for Repository pattern."""

    def __init__(self) -> None:
        self.seen = []

    seen: MutableSequence[TModel_contra]


class SyncEventstoreAbstractRepository(abc.ABC):
    def __init__(self, publisher: Optional[SyncEventstreamPublisher] = None) -> None:
        self.publisher = publisher
        self.stream_buffer: MutableSequence[Message[Any]] = []

    @abc.abstractmethod
    def _add(self, message: Message[Any]) -> None:
        """
        Add a message to the storage backend of event repository.
        """

    def add(self, message: Message[Any]) -> None:
        """
        Add the message to the storage backend and mark as seen

        seen message will be sent to the eventstream only if the unit of work
        has properly commit the transaction.
        If the transaction is rollback, then, message will be dropped too from the
        eventstream.
        """
        self._add(message)
        self.stream_buffer.append(message)

    def publish_eventstream(self) -> None:
        """
        Publish seen message to the eventstream.
        """
        stream_buffer, self.stream_buffer = self.stream_buffer, []
        if not self.publisher:
            return

        for message in stream_buffer:
            self.publisher.send_message(message)


class SyncSinkholeEventstoreRepository(SyncEventstoreAbstractRepository):
    """An eventstore that drop all the message."""

    def _add(self, message: Message[Any]) -> None:
        """Do nothing. The sinkhole drop every message."""



"""
Propagate commands and events to every registered handles.

"""

import importlib
import inspect
import logging
from collections import defaultdict
from typing import Any, Generic, cast

import venusian  # type: ignore

from jeepito.domain.model import Command, Event, Message
from jeepito.typing import SyncMessageHandler, TMessage, TSyncUow

from .unit_of_work import SyncUnitOfWorkTransaction, TRepositories

log = logging.getLogger(__name__)
VENUSIAN_CATEGORY = "jeepito"


class ConfigurationError(RuntimeError):
    """Prevents bad usage of the add_listener."""


def sync_listen(
    wrapped: SyncMessageHandler[TMessage, TSyncUow],
) -> SyncMessageHandler[TMessage, TSyncUow]:
    """
    Decorator to listen for a command or an event.

    Note that you can handle one listener for a command, and many for events.
    The command handler result is returned by the handle call of the message bus.
    """

    def callback(
        scanner: venusian.Scanner,
        name: str,
        ob: SyncMessageHandler[TMessage, TSyncUow],
    ) -> None:
        if not hasattr(scanner, VENUSIAN_CATEGORY):
            return  # coverage: ignore
        argsspec = inspect.getfullargspec(ob)
        msg_type = argsspec.annotations[argsspec.args[0]]
        scanner.jeepito.add_listener(msg_type, wrapped)  # type: ignore

    venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
    return wrapped


class SyncMessageBus(Generic[TRepositories]):
    """Store all the handlers for commands an events."""

    def __init__(self) -> None:
        self.commands_registry: dict[
            type[Command[Any]], SyncMessageHandler[Command[Any], Any]
        ] = {}
        self.events_registry: dict[
            type[Event[Any]], list[SyncMessageHandler[Event[Any], Any]]
        ] = defaultdict(list)

    def add_listener(
        self, msg_type: type[Message[Any]], callback: SyncMessageHandler[Any, Any]
    ) -> None:
        if issubclass(msg_type, Command):
            if msg_type in self.commands_registry:
                raise ConfigurationError(
                    f"{msg_type} command has been registered twice"
                )
            self.commands_registry[msg_type] = callback
        elif issubclass(msg_type, Event):
            self.events_registry[msg_type].append(callback)
        else:
            raise ConfigurationError(
                f"Invalid usage of the listen decorator: "
                f"type {msg_type} should be a command or an event"
            )

    def remove_listener(
        self, msg_type: type, callback: SyncMessageHandler[Any, Any]
    ) -> None:
        if issubclass(msg_type, Command):
            if msg_type not in self.commands_registry:
                raise ConfigurationError(f"{msg_type} command has not been registered")
            del self.commands_registry[msg_type]
        elif issubclass(msg_type, Event):
            try:
                self.events_registry[msg_type].remove(callback)
            except ValueError as exc:
                raise ConfigurationError(
                    f"{msg_type} event has not been registered"
                ) from exc
        else:
            raise ConfigurationError(
                f"Invalid usage of the listen decorator: "
                f"type {msg_type} should be a command or an event"
            )

    def handle(
        self, message: Message[Any], uow: SyncUnitOfWorkTransaction[TRepositories]
    ) -> Any:
        """
        Notify listener of that event registered with `jeepito.add_listener`.
        Return the first event from the command.
        """
        queue = [message]
        idx = 0
        ret = None
        while queue:
            message = queue.pop(0)
            if not isinstance(message, (Command, Event)):
                raise RuntimeError(f"{message} was not an Event or Command")
            msg_type = type(message)
            if msg_type in self.commands_registry:
                cmdret = self.commands_registry[msg_type](  # type: ignore
                    cast(Command[Any], message), uow
                )
                if idx == 0:
                    ret = cmdret
                queue.extend(uow.uow.collect_new_events())
            elif msg_type in self.events_registry:
                for callback in self.events_registry[msg_type]:  # type: ignore
                    callback(cast(Event[Any], message), uow)
                    queue.extend(uow.uow.collect_new_events())
            uow.eventstore.add(message)
            idx += 1
        return ret

    def scan(
        self,
        *mods: str,
    ) -> None:
        """
        Scan the module (or modules) containing service handlers.

        when a message is handled by the bus, the bus propagate the message
        to hook functions, called :term:`Service Handler` that receive the message,
        and a :term:`Unit Of Work` to process it has a business transaction.
        """
        scanner = venusian.Scanner(jeepito=self)
        for modname in mods:
            if modname.startswith("."):
                raise ValueError(
                    f"scan error: relative package unsupported for {modname}"
                )
            mod = importlib.import_module(modname)
            scanner.scan(mod, categories=[VENUSIAN_CATEGORY])  # type: ignore



"""
Jeepito API.
"""

try:
    # add an ignore for python 3.7 which does not know this standard library
    from importlib.metadata import version  # type: ignore
except ImportError:  # coverage: ignore
    from pkg_resources import get_distribution  # type: ignore

    # pythond 3.7 fallback
    def version(distribution_name: str) -> str:
        return get_distribution(distribution_name).version


from pydantic import Field

from .domain.model import Command, Event, Message, Metadata, Model
from .service._async.eventstream import (
    AsyncAbstractEventstreamTransport,
    AsyncEventstreamPublisher,
    AsyncSinkholeEventstreamTransport,
)
from .service._async.registry import AsyncMessageBus, async_listen
from .service._async.repository import AsyncAbstractRepository
from .service._async.unit_of_work import (
    AsyncAbstractUnitOfWork,
    AsyncEventstoreAbstractRepository,
    AsyncSinkholeEventstoreRepository,
    AsyncUnitOfWorkTransaction,
)
from .service._sync.eventstream import (
    SyncAbstractEventstreamTransport,
    SyncEventstreamPublisher,
    SyncSinkholeEventstreamTransport,
)
from .service._sync.registry import SyncMessageBus, sync_listen
from .service._sync.repository import SyncAbstractRepository
from .service._sync.unit_of_work import (
    SyncAbstractUnitOfWork,
    SyncEventstoreAbstractRepository,
    SyncSinkholeEventstoreRepository,
    SyncUnitOfWorkTransaction,
)
from .service.eventstream import AbstractMessageSerializer

__version__ = version("Jeepito")

__all__ = [
    # models
    "Model",
    "Field",
    "Message",
    "Metadata",
    "Command",
    "Event",
    # Repository
    "AsyncAbstractRepository",
    "SyncAbstractRepository",
    # Unit of work
    "AsyncAbstractUnitOfWork",
    "AsyncUnitOfWorkTransaction",
    "SyncAbstractUnitOfWork",
    "SyncUnitOfWorkTransaction",
    "AsyncEventstoreAbstractRepository",
    "SyncEventstoreAbstractRepository",
    "AsyncSinkholeEventstoreRepository",
    "SyncSinkholeEventstoreRepository",
    # Registry
    "async_listen",
    "sync_listen",
    "AsyncMessageBus",
    "SyncMessageBus",
    # Eventstream
    "AbstractMessageSerializer",
    "AsyncAbstractEventstreamTransport",
    "AsyncSinkholeEventstreamTransport",
    "AsyncEventstreamPublisher",
    "SyncAbstractEventstreamTransport",
    "SyncSinkholeEventstreamTransport",
    "SyncEventstreamPublisher",
]



import abc
from collections.abc import Mapping
from typing import Any

from jeepito.domain.model import Message
from jeepito.service.eventstream import AbstractMessageSerializer, MessageSerializer


class AsyncAbstractEventstreamTransport(abc.ABC):
    """
    Transport a message to the event stream.
    """

    @abc.abstractmethod
    async def send_message_serialized(self, message: Mapping[str, Any]) -> None:
        """Publish a serialized message to the eventstream."""


class AsyncSinkholeEventstreamTransport(AsyncAbstractEventstreamTransport):
    """
    Drop all messages.

    By default, the events are not streamed until it is configured to do so.
    """

    async def send_message_serialized(self, message: Mapping[str, Any]) -> None:
        """Do nothing."""


default_serializer = MessageSerializer()


class AsyncEventstreamPublisher:
    """
    Publish a message to the event stream.

    :param serializer: Used to serialize the Message
    :param transport: Used to send the serialized message to the eventstream.
    """

    def __init__(
        self,
        transport: AsyncAbstractEventstreamTransport,
        serializer: AbstractMessageSerializer = default_serializer,
    ) -> None:
        """Publish a message to the eventstream."""
        self.transport = transport
        self.serializer = serializer

    async def send_message(self, message: Message[Any]) -> None:
        """
        Publish a message to the eventstream.

        To publish a message in the eventstream, the flag parameter "published" of
        the metadata of the message must be set to true.
        By default, message are not pushed to the queue, given the control of
        private message, such as command, and public event, shared with eventstream
        consumers.
        """
        if not message.metadata.published:
            return
        evt = self.serializer.serialize_message(message)
        await self.transport.send_message_serialized(evt)
